#include <linux/interrupt.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/cdev.h>
#include <linux/fs.h>

#include <asm/uaccess.h>


static dev_t          dev_reveil;
static struct cdev    cdev_reveil;
static struct class * class_reveil = NULL;

static int read_reveil (struct file * filp, char * buffer,
                          size_t length, loff_t * offset);

struct timer_list timer_reveil;
static void timer_function(unsigned long);

static struct file_operations fops_reveil = {
	.owner   =  THIS_MODULE,
	.read    =  read_reveil,
};

struct timespec ts_reveil;
static DECLARE_WAIT_QUEUE_HEAD(wq_reveil);


static int __init init_reveil (void)
{
	int erreur;
	
	erreur = alloc_chrdev_region(& dev_reveil, 0, 1, THIS_MODULE->name);
	if (erreur < 0)
		return erreur;
		
	class_reveil = class_create(THIS_MODULE, "classe_reveil");
	if (IS_ERR(class_reveil)) {
		unregister_chrdev_region(dev_reveil, 1);
		return -EINVAL;
	}
	device_create(class_reveil, NULL, dev_reveil,
	              NULL, THIS_MODULE->name);

	cdev_init(& cdev_reveil, & fops_reveil);

	erreur = cdev_add(& cdev_reveil, dev_reveil, 1);
	if (erreur != 0) {
		device_destroy(class_reveil, dev_reveil);
		class_destroy(class_reveil);
		unregister_chrdev_region(dev_reveil, 1);
		return erreur;
	}
	init_timer(& timer_reveil);
	timer_reveil.function = timer_function;
	timer_reveil.expires  = HZ;
	add_timer(& timer_reveil);
	return 0; 
}



static void __exit exit_reveil (void)
{
	del_timer(& timer_reveil);
	cdev_del(& cdev_reveil);
	device_destroy(class_reveil, dev_reveil);
	class_destroy(class_reveil);
	unregister_chrdev_region(dev_reveil, 1);
}


static int read_reveil(struct file * filp, char * buffer,
                        size_t length, loff_t * offset)
{
	char chaine[80];

	ts_reveil.tv_sec = 0;
	if (wait_event_interruptible(wq_reveil, (ts_reveil.tv_sec != 0)) != 0)
			return -ERESTARTSYS;
	sprintf(chaine, "%ld %ld\n", ts_reveil.tv_sec, ts_reveil.tv_nsec);
	if(copy_to_user(buffer, chaine, strlen(chaine)+1) != 0)
		return -EFAULT;
	return strlen(chaine)+1;
}



static void timer_function(unsigned long unused)
{
	ktime_get_real_ts(& ts_reveil);
	wake_up_interruptible(& wq_reveil);
	mod_timer(& timer_reveil, jiffies + HZ/100);
}


module_init(init_reveil);
module_exit(exit_reveil);
MODULE_LICENSE("GPL");

